package ssslib;
require _ssslib;
require strict;

sub new {
    my $class = shift;
    my ($conffile) = $_;
    my $self = bless {sss => _ssslib::SSS_Init($conffile, 1 )};
    return $self;
}

sub DESTROY {
    my $self = shift;
    my $sss = $self->{sss};
    if($sss) {
	_ssslib::SSS_Close($sss);
    }
}

sub SDRegister {
    my $self = shift;
    my $sss = $self->{sss};
    my ($service, $host, $port, $proto, $schema, $tier) = @_;
    my $retval = _ssslib::SSS_SDRegister($sss, $service, $host, $port, $proto, $schema, $tier);

    if ((defined($retval)) and ($retval == -1)) {
	my $errno = _ssslib::SSS_t_sss_errno_get($sss);
	print "DEBUG: \$errno=$errno\n";
	throw_exception($errno);
    }

    return $retval;
}

sub SDRemove {
    my $self = shift;
    $sss = $self->{sss};
    my ($service, $proto) = @_;
    my $retval = _ssslib::SSS_SDRemove($sss, $service, $proto);

    if ((defined($retval)) and ($retval == -1)) {
	my $errno = _ssslib::SSS_t_sss_errno_get($sss);
	print "DEBUG: \$errno=$errno\n";
	throw_exception($errno);
    }

    return $retval;
}

sub ServerInit {
    my $self = shift;
    $sss = $self->{sss};
    my ($socket) = @_;
    my $retval = _ssslib::SSS_ServerInit($sss, $socket);

    if (!$retval) {
	my $errno = _ssslib::SSS_t_sss_errno_get($sss);
	print "DEBUG: \$errno=$errno\n";
	throw_exception($errno);
    }

    return $retval;
}

sub ClientInit {
    my $self = shift;
    $sss = $self->{sss};
    my ($service) = @_;
    my $retval = _ssslib::SSS_ClientInit($sss, $service);

    if (!$retval) {
	my $errno = _ssslib::SSS_t_sss_errno_get($sss);
	print "DEBUG: \$errno=$errno\n";
	throw_exception($errno);
    }

    return $retval;
}

sub ClientClose {
    my $self = shift;
    $sss = $self->{sss};
    my ($handle) = @_;
    my $retval = _ssslib::SSS_ClientClose($sss, $handle);

    if ((defined($retval)) and ($retval == -1)) {
	my $errno = _ssslib::SSS_t_sss_errno_get($sss);
	print "DEBUG: \$errno=$errno\n";
	throw_exception($errno);
    }

    return $retval;
}

sub ServerClose {
    my $self = shift;
    $sss = $self->{sss};
    $self = shift;
    my ($handle) = @_;
    my $retval = _ssslib::SSS_ServerClose($sss, $handle);

    if ((defined($retval)) and ($retval == -1)) {
	my $errno = _ssslib::SSS_t_sss_errno_get($sss);
	print "DEBUG: \$errno=$errno\n";
	throw_exception($errno);
    }

    return $retval;
}

sub SendMessage {
    my $self = shift;
    $sss = $self->{sss};
    my ($handle, $message) = @_;
    my $retval = _ssslib::SSS_SendMessage($sss, $handle, $message);

    if ((defined($retval)) and ($retval == -1)) {
	my $errno = _ssslib::SSS_t_sss_errno_get($sss);
	print "DEBUG: \$errno=$errno\n";
	throw_exception($errno);
    }

    return $retval;
}

sub RecvMessage {
    my $self = shift;
    $sss = $self->{sss};
    my ($handle) = @_;
    my $retval = _ssslib::SSS_RecvMessage($sss, $handle);

    if ((defined($retval)) and ($retval == -1)) {
	my $errno = _ssslib::SSS_t_sss_errno_get($sss);
	print "DEBUG: \$errno=$errno\n";
	throw_exception($errno);
    }

    return $retval;
}

sub throw_exception {
    my $self = shift;
    my ($errnum) = @_;

    if($errnum == $_ssslib::SSS_XML_PARSE) { die "Error parsing XML\n"; }
    elsif($errnum == $_ssslib::SSS_SD_FAIL ) { die "Error in communication with service directory\n"; }
    elsif($errnum == $_ssslib::SSS_PROT_NAME_LONG ) { die "Protocol name longer than 124 characters\n"; }
    elsif($errnum == $_ssslib::SSS_NO_SUCH_CONN ) { die "A connection by that handle does not exist\n"; }
    elsif($errnum == $_ssslib::SSS_PROTO_NO_SUPPORT ) { die "Unsupported protocol\n"; }
    elsif($errnum == $_ssslib::SSS_FAM_NO_SUPPORT ) { die "Unsupported family\n"; }
    elsif($errnum == $_ssslib::SSS_NOMEM ) { die "Out of memory\n"; }
    elsif($errnum == $_ssslib::SSS_NOSOCKMEM ) { die "No memory to make a new socket\n"; }
    elsif($errnum == $_ssslib::SSS_INVALPROT ) { die "Invalid protocol\n"; }
    elsif($errnum == $_ssslib::SSS_HOST_NOT_FOUND ) { die "Host not found\n"; }
    elsif($errnum == $_ssslib::SSS_NO_ADDRESS ) { die "No address given\n"; }
    elsif($errnum == $_ssslib::SSS_UNRECOVERABLE_DNS_ERROR ) { die "Unrecoverable dns error\n"; }
    elsif($errnum == $_ssslib::SSS_TRY_AGAIN ) { die "Temporary DNS failure - try again\n"; }
    elsif($errnum == $_ssslib::SSS_BAD_FD ) { die "Bad file descriptor\n"; }
    elsif($errnum == $_ssslib::SSS_FAULT ) { die "Memory fault\n"; }
    elsif($errnum == $_ssslib::SSS_ALREADY_CONNECTED ) { die "Socket already connected\n"; }
    elsif($errnum == $_ssslib::SSS_CONN_REFUSED ) { die "Connection refused\n"; }
    elsif($errnum == $_ssslib::SSS_TIMEOUT ) { die "Connection timed out\n"; }
    elsif($errnum == $_ssslib::SSS_NET_UNREACH ) { die "Network unreachable\n"; }
    elsif($errnum == $_ssslib::SSS_ADDR_IN_USE ) { die "Address already in use\n"; }
    elsif($errnum == $_ssslib::SSS_NO_FREE_PORTS ) { die "No free ports\n"; }
    elsif($errnum == $_ssslib::SSS_NO_CONNECT_PERM ) { die "No permissions to connect\n"; }
    elsif($errnum == $_ssslib::SSS_NOSOCKPERM ) { die "No permissions to make socket\n"; }
    elsif($errnum == $_ssslib::SSS_NOT_A_SOCKET ) { die "FD passed in is not a socket\n"; }
    elsif($errnum == $_ssslib::SSS_SOCK_NOT_CONNECTED ) { die "Socket is not connected\n"; }
    elsif($errnum == $_ssslib::SSS_INVALID_SIZE ) { die "Invalid size\n"; }
    elsif($errnum == $_ssslib::SSS_INTERRUPT ) { die "Interrupt\n"; }
    elsif($errnum == $_ssslib::SSS_INVAL_ARG ) { die "Invalid argument\n"; }
    elsif($errnum == $_ssslib::SSS_BROKEN_PIPE ) { die "Broken pipe\n"; }
    elsif($errnum == $_ssslib::SSS_AUTH_FAILED ) { die "Authentication error\n"; }
    elsif($errnum == $_ssslib::SSS_LIBRARY_ERROR ) { die "Error loading protocol library\n"; }
    elsif($errnum == $_ssslib::SSS_EOF ) { die "EOF on read\n"; }
    elsif($errnum == $_ssslib::SSS_NEED_MORE_INFO ) { die "Protocol needs more configuration information\n"; }
    elsif($errnum == $_ssslib::SSS_INIT_ERROR ) { die "Error in init.\n"; }
    elsif($errnum == $_ssslib::SSS_PARSE_ERROR ) { die "Parse Error.\n"; }
    elsif($errnum == $_ssslib::SSS_PROTOCOL_ERROR ) { die "Protocol Error.\n"; }
    elsif($errnum == $_ssslib::SSS_SOCKET_ERROR ) { die "Socket error.\n"; }
    elsif($errnum == $_ssslib::SSS_CONNECT_ERROR ) { die "Error trying to connect.\n"; }
    elsif($errnum == $_ssslib::SSS_COMMUNICATION_ERROR ) { die "Communication error.\n"; }
    elsif($errnum == $_ssslib::SSS_INTERNAL_ERROR ) { die "Internal error.\n"; }
    else { die "error number $errnum not handled!\n"; }

}

1;
