#!/usr/bin/env python

from os import popen3, system, unlink, dup2, execl, wait, stat
from select import error as selecterror
import sys
from sys import exit
from sss.node import Node, Msgbuf
from sss.server import SSSSelectServer, AddEvent
from sss.ssslib import ConnectError
from string import join,split
from syslog import syslog, LOG_INFO, LOG_ERR, LOG_LOCAL0, openlog
from tempfile import mktemp
from xml.dom.minidom import Document
from time import time
from select import select

mpdpath='/usr/mpi/bin'
outputspool='/tmp'

class Arg(Node):
    __attrfields__=['idx','value']

class Env(Node):
    __attrfields__=['name','value']    

class HostSpec(Msgbuf):
    pass

class ProcessSpec(Node):
    __attrfields__=['cwd','exec','path','user']
    typemap={'arg':Arg,'env':Env,'host-spec':HostSpec}

class ProcessGroup(Node):
    __attrfields__=['submitter','output','totalprocs','state']
    typemap={'process-spec':ProcessSpec, 'host-spec':HostSpec}

    def __setdefaults__(self):
        setattr(self,'host-spec','')
        setattr(self,'mpdpid',-1)
        for attr in ['outfile','errfile','mpdfile']:
            setattr(self,attr,mktemp())

    def MPIStart(self):
        self.tagname="process-group"
        self.__attrfields__.append('pgid')
        self.state="STARTED"
        syslog(LOG_INFO,"PGid %s Started"%(self.pgid))
        try:
            AddEvent('process-manager','process_start',self.pgid)
        except:
            syslog(LOG_ERR, "Error sending process_start for pgid %s"%(self.pgid))
        # Dump out xml file
        self.tagname="create-process-group"
        open(self.mpdfile,'w').write(self.toxmldom().toxml())
        self.tagname="process-group"
        syslog(LOG_INFO, "running :%s:"%("%s/mpdrun -f %s"%(mpdpath,self.mpdfile)))
        pid=fork()
        if pid:
            self.mpdpid=pid
        else:
            null=open('/dev/null','r')
            out=open(self.outfile,'w')
            err=open(self.errfile,'w')
            dup2(null.fileno(),sys.__stdin__.fileno())
            dup2(out.fileno(),sys.__stdout__.fileno())
            dup2(err.fileno(),sys.__stderr__.fileno())
            execl("%s/mpdrun"%(mpdpath), 'mpdrun','-f', self.mpdfile)

    def MPIFinish(self):
        self.output = open(self.outfile).read()
        self.error  = open(self.errfile).read()
        syslog(LOG_INFO,"output was :%s:"%self.output)
        syslog(LOG_INFO,"error was :%s:"%self.error)
        for file in ['outfile','errfile','mpdfile']:
            unlink(getattr(self,file))
        syslog(LOG_INFO,"PGid %s Exited"%(self.pgid))
        try:
            AddEvent('process-manager','process_end',self.pgid)
        except:
            syslog(LOG_ERR, "Error sending process_end for pgid %s"%(self.pgid))
        # Send out completion
        self.state="FINISHED"

    def Signal(self,signal):
        system("%s/mpdsigjob %s -a %s"%(mpdpath,signal,self.pgid))
        syslog(LOG_INFO,"PGid %s signaled with %s"%(self.pgid,signal))

    def toxmldom(self,spec=None):
        initdom=Node.toxmldom(self,spec)
        doc=Document()
        if hasattr(self,"output"):
            if self.output:
                out=doc.createElement("output")
                data=doc.createTextNode(self.output)
                out.appendChild(data)
                initdom.appendChild(out)
        if hasattr(self,"error"):
            if self.error:
                err=doc.createElement("error")
                data=doc.createTextNode(self.error)
                err.appendChild(data)
                initdom.appendChild(err)
        return initdom

class P4_ProcessGroup(Node):
    __attrfields__=['submitter','output','totalprocs']
    __textfields__=['host-spec']
    typemap={'process-spec':ProcessSpec}

    def MPIRun(self):
        # Dump out xml file
        tempfile=mktemp()
        syslog(LOG_INFO,"my host-spec is :%s:"%getattr(self,"host-spec"))
        open(tempfile,'w').write(getattr(self,"host-spec"))
        self.tagname="process-group"
        cmdline = 'su - %s -c "%s/mpirun -nolocal -np %s -machinefile %s'%(self.submitter,mpdpath,self.totalprocs,tempfile)
        cmdline += ' %s '%getattr(self.children[0],"exec")
        for arg in filter(lambda x:isinstance(x,Arg),self.children[0].children):
            cmdline += ' %s '%arg.value
        cmdline += '"'
        syslog(LOG_INFO, "running :%s:"%cmdline)
        (foo,out,err)=popen3(cmdline)
        #unlink(tempfile)
        self.output = ""
        self.error  = ""
        out_should_exit = 0
        err_should_exit = 0
        while 1:
            socklist = []
            if not out_should_exit:
                socklist.append(out)
            if not err_should_exit:
                socklist.append(err)
            [reads,writes,excepts] = select(socklist,[],[])
            for r in reads:
                str = r.read(1)
                if r.fileno() == out.fileno():
                    self.output += str
                    if not len(str):
                        out_should_exit = 1
                if r.fileno() == err.fileno():
                    self.error += str
                    if not len(str):
                        err_should_exit = 1
            if out_should_exit and err_should_exit:
                syslog(LOG_INFO,"output was :%s:"%self.output)
                syslog(LOG_INFO,"error was :%s:"%self.error)
                return
 
    def Run(self):
        self.tagname="process-group"
        self.__attrfields__.append('pgid')
        self.state="STARTED"
        syslog(LOG_INFO,"PGid %s Started"%(self.pgid))
        try:
            AddEvent('process-manager','process_start',self.pgid)
        except ConnectError:
            syslog(LOG_ERR, "Error sending process_start for pgid %s"%(self.pgid))
        # MPIRun
        self.MPIRun()
        syslog(LOG_INFO,"PGid %s Exited"%(self.pgid))
        try:
            AddEvent('process-manager','process_end',self.pgid)
        except ConnectError:
            syslog(LOG_ERR, "Error sending process_end for pgid %s"%(self.pgid))
        # Send out completion
        self.state="FINISHED"
        return None
    
    def Signal(self,signal):
        pass
        #system("%s/mpdsigjob %s -a %s"%(mpdpath,signal,self.pgid))
        #syslog(LOG_INFO,"PGid %s signaled with %s"%(self.pgid,signal))

    def toxmldom(self,spec=None):
        initdom=Node.toxmldom(self,spec)
        doc=Document()
        if hasattr(self,"output"):
            out=doc.createElement("output")
            data=doc.createTextNode(self.output)
            out.appendChild(data)
            initdom.appendChild(out)
        if hasattr(self,"error"):
            err=doc.createElement("error")
            data=doc.createTextNode(self.error)
            err.appendChild(data)
            initdom.appendChild(err)
        return initdom

class mpd(SSSSelectServer):
    __implementation__='mpdpm'
    __component__='process-manager'
    
    def __parseinput__(self):
        self.validate=0
        self.dispatch={"create-process-group":self.HandleXMLCreatePG,
                       "signal-process-group":self.HandleXMLPGSignal,
                       "kill-process-group":self.HandleXMLPGKill,
                       "get-process-group-info":self.HandleXMLPGInfo}
        self.pg={}
        self.e=[]
        try:
            stat("%s/mpdrun"%(mpdpath))
        except OSError:
            openlog('mpdpm',0,LOG_LOCAL0)
            syslog(LOG_ERR, "Error finding mpd. exitting")
            exit(1)

    def GetPGID(self):
        return str(time())

    def ServeForever(self):
        while not self.shut:
            try:
                self.Serve()
                # handle completions
                for pid in self.e:
                    pg=filter(lambda x:x.mpdpid==pid,self.pg.values())
                    if len(pg) == 0:
                        syslog(LOG_ERR, "Data consistency error for pid %s"%(pid))
                    elif len(pg) == 1:
                        pg[0].MPIFinish()
                        self.e.remove(pid)
            except selecterror:
                pass

    def SigChildHand(self,sig,frame):
        (pid,stat)=wait()
        self.e.append(pid)

    def HandleXMLCreatePG(self,xe,(peer,port)):
        # xe looks like
        # <create-process-group submitter='desai' pgid='1' totalprocs='32' output='discard'>
        #  <process-spec exec='/bin/foo' cwd='/etc' path='/bin:/usr/sbin' range='1-32'
        #              co-process='tv-server'>
        #     <arg idx='1' value='-v'/>
        #  </process-spec>
        #  <host-spec>
        #    node1
        #    node2
        #  </host-spec>
        # </create-process-group>
        if '-1' in argv:
            newpg=P4_ProcessGroup(xe)
        else:
            newpg=ProcessGroup(xe)
        newpg.pgid=self.GetPGID()
        self.pg[newpg.pgid]=newpg
        syslog(LOG_INFO, "proc-group %s created for %s"%(newpg.pgid,newpg.submitter))
        self.pg[newpg.pgid].MPIStart()
        # Deal with stdio
        creator=Document()
        response=creator.createElement("process-group")
        response.setAttribute('pgid',newpg.pgid)
        return response

    def HandleXMLPGSignal(self,xe,(peer,port)):
        sig=xe.getAttribute('signal')
        response=Document().createElement("process-groups")
        for pgrp in xe.getElementsByTagName('process-group'):
            pgid=pgrp.getAttribute('pgid')
            if pgid in self.pg.keys():
                self.pg[pgid].Signal(sig)
            else:
                syslog(LOG_INFO,"got stale pgid %s"%pgid)
            doc=Document().createElement("process-group")
            doc.setAttribute('pgid',pgid)
            response.appendChild(doc)
        return response

    def HandleXMLPGKill(self,xe,(peer,port)):
        response=Document().createElement("process-groups")
        for pgrp in xe.getElementsByTagName('process-group'):
            pgid=pgrp.getAttribute('pgid')
            if pgid in self.pg.keys():
                self.pg[pgid].Signal("SIGKILL")
            else:
                syslog(LOG_INFO,"got stale pgid %s"%pgid)
            doc=Document().createElement("process-group")
            doc.setAttribute('pgid',pgid)
            response.appendChild(doc)
        return response

    def HandleXMLPGInfo(self,xe,(peer,port)):
        creator=Document()
        response=creator.createElement("process-groups")
        # process xe
        for node in xe.childNodes:
            matches=filter(lambda x:x.Match(Node(node)), self.pg.values())
            map(lambda x:response.appendChild(x.toxmldom(Node(node))), matches)
        return response

if __name__ == '__main__':
    from getopt import getopt, GetoptError
    from os import fork
    from sys import argv,exit
    try:
        (opts,args)=getopt(argv[1:],'1dm:',['daemon='])
    except GetoptError,msg:
        print "%s\nUsage:\nmpdpm.py [-1] [-d] [-m <path to mpd> ] [--daemon <pidfile>]"%(msg)
        exit(1)
    daemon=filter(lambda x:(x[0] == '--daemon'),opts)
    mpath=filter(lambda x:(x[0] == '-m'),opts)
    ldebug=len(filter(lambda x:(x[0] == '-d'),opts))
    if mpath:
        mpdpath=mpath[0][1]
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0][1])
    s=mpd(debug=ldebug)
    s.ServeForever()
