#include<ssslib.h>
#include<stdlib.h>
#include<openssl/md5.h>

struct auth_challenge_state {
  int sss_errno;
  int debug;
  struct transport *transport;
  char *password;
};

void *auth_challenge_LTX_auth_getstate(SSS_ConfigDict cfg,int debug,void *transport){
  struct auth_challenge_state *ret;

  if (debug)
    printf("auth_challenge: starting init\n");

  ret = malloc(sizeof(struct auth_challenge_state));

  ret->transport = transport;
  ret->debug = debug;
  if (!cfg){
    if (debug){
      printf("auth_challenge: need a \"password=\" in configuration file for this protocol.\n");
      free(ret);
      return NULL;
    }
  }
  ret->password = SSS_GetFromConfigDict(cfg,"password");
  if (!ret->password){
    if (debug){
      printf("auth_challenge: need a \"password=\" in configuration file for this protocol.\n");
      free(ret);
      return NULL;
    }
  }

  if (debug)
    printf("auth_challenge: successful init\n");

  return ret;
}

int auth_challenge_LTX_cli_auth(void *auth_state){
  char buf[4096];
  char response[4096];
  unsigned char md5sum[4096];
  char authchar[1];
  char smallbuf[3];
  char md5buf[34];
  unsigned char *tmp;
  int i;
  int rc;
  struct auth_challenge_state *state = auth_state;

  authchar[0] = '3';

  bzero(buf,4096);
  bzero(response,4096);
  bzero(md5sum,4096);

  rc = state->transport->read(state->transport->state,response,4096);
  if (rc == -1){
    memcpy(&(state->sss_errno),state->transport->state,sizeof(int));
    return -1;
  }
  
  response[strlen(response) - 1] = '\0';
  sprintf(buf,"%s%s",response,state->password);
  tmp = MD5((unsigned char *)buf,(unsigned long)strlen(buf),md5sum);
  md5buf[0] = '\0';
  for(i=0;i<16;i++){
    sprintf(smallbuf,"%02x",md5sum[i]);
    strcat(md5buf,smallbuf);
  }

  sprintf(md5buf,"%s\n",md5buf);
  rc = state->transport->writen(state->transport->state,md5buf,strlen(md5buf));
  if (rc == -1){
    memcpy(&(state->sss_errno),state->transport->state,sizeof(int));
    return -1;
  }

  rc = state->transport->readn(state->transport->state,authchar,1);
  if (rc == -1){
    memcpy(&(state->sss_errno),state->transport->state,sizeof(int));
    return -1;
  }
    
  if (authchar[0] == '0'){
    if (state->debug)
      printf("auth_challenge: authentication failed\n");
    state->sss_errno = SSS_AUTH_FAILED;
    return -1;
  }
  
  if (state->debug)
    printf("auth_challenge: authentication successful\n");
  return 0;
}

int auth_challenge_LTX_serv_auth(void *auth_state){
  long randnum;
  char buf[4096];
  char response[4096];
  unsigned char md5sum[4096];
  char smallbuf[3];
  char md5buf[34];
  unsigned char *tmp;
  int i,rc;
  char tmpchr = '0';
  struct auth_challenge_state *state = auth_state;

  srand(time(NULL));
  randnum = random() % 10000000;
  
  bzero(buf,4096);
  bzero(response,4096);
  bzero(md5sum,4096);

  sprintf(response,"%li\n",randnum);
  rc = state->transport->writen(state->transport->state,response,strlen(response));
  if (rc == -1){
    memcpy(&(state->sss_errno),state->transport->state,sizeof(int));
    return -1;
  }
  
  bzero(response,4096);

  while (tmpchr != '\n'){
    rc = state->transport->readn(state->transport->state,&tmpchr,1);
    if (rc == -1){
      memcpy(&(state->sss_errno),state->transport->state,sizeof(int));
      return -1;
    }
    strncat(response,&tmpchr,1);
  }

  response[strlen(response)-1] = '\0';
  sprintf(buf,"%li%s",randnum,state->password);
  tmp = MD5((unsigned char *)buf,(unsigned long)strlen(buf),md5sum);
  md5buf[0] = '\0';
  for (i=0;i<16;i++){
    sprintf(smallbuf,"%02x",md5sum[i]);
    strcat(md5buf,smallbuf);
  }

  if (strcmp(response,md5buf)){
    if (state->debug)
      printf("auth_challenge: authentication failed\n");
    tmpchr = '0';
    state->transport->writen(state->transport->state,&tmpchr,1);
    state->sss_errno = SSS_AUTH_FAILED;
    return -1;
  }

  tmpchr = '1';
  state->transport->writen(state->transport->state,&tmpchr,1);
  if (state->debug)
    printf("auth_challenge: authentication successful\n");
  return 0;
}

int auth_challenge_LTX_auth_close(void *auth_state){
  int debug;
  struct auth_challenge_state *state = auth_state;
  debug = state->debug;
  free(auth_state);
  if (debug)
    printf("auth_challenge: successful close\n");
  return 0;
}

