#include <ssslib.h>		/* for ssslib */
#include <ltdl.h>		/* for modules */
#include <sys/types.h>		/* for fds */
#include <sys/socket.h>		/* for sockets */
#include <sys/stat.h>		/* for fds */
#include <fcntl.h>		/* for open/close/read/etc */
#include <unistd.h>		/* for stat */
#include <stdlib.h>		/* for malloc/free */
#include <string.h>		/* for string ops */
#include <stdio.h>		/* for printf */
#include <netdb.h>		/* for gethostbyname */
#include <netinet/in.h>		/* for sockaddr, htons etc */
#include <arpa/inet.h>		/* for inet_addr */
#include <errno.h>		/* for errno */
#include <time.h>		/* for time */

#include <config.h>

char buf[1];

char *my_strdup(char *s){
  char *ret = malloc(strlen(s)+1);
  strcpy(ret,s);
  return ret;
}

int module_load_funcs(int debug,
		      char *protocol,
		      int (**cli_connect)(struct SSS_Protocol *,char *,int),
		      int (**send_message)(struct SSS_Protocol *,char *),
		      char *(**recv_message)(struct SSS_Protocol *),
		      struct SSS_Protocol *(**get_state)(SSS_ConfigDict,int,int *),
		      int (**close_conn)(struct SSS_Protocol *),
		      int (**serv_connect)(struct SSS_Protocol *,int),
		      void **handle){
  lt_dlhandle hand;
  char *modstr;
  int modstr_len;

  modstr_len = snprintf(buf,1,"%s/%s",MODPATH,protocol);
  modstr = malloc(modstr_len+1);
  sprintf(modstr,"%s/%s",MODPATH,protocol);

  hand = lt_dlopenext(modstr);

  if (!hand && debug){
    fprintf(stderr,"loading %s failed\n",modstr);
    return -1;
  }

  *cli_connect = (int (*)(struct SSS_Protocol *,char *,int))lt_dlsym(hand,"cli_connect");
  if(!(*cli_connect) && debug){
    fprintf(stderr,"can't resolve cli_connect\n");
    return -1;
  }

  *send_message = (int (*)(struct SSS_Protocol *,char *))lt_dlsym(hand,"send_message");
  if(!(*send_message) && debug){
    fprintf(stderr,"can't resolve send_message\n");
    return -1;
  }

  *recv_message = (char *(*)(struct SSS_Protocol *))lt_dlsym(hand,"recv_message");
  if(!(*recv_message) && debug){
    fprintf(stderr,"can't resolve recv_message\n");
    return -1;
  }

  *get_state = (struct SSS_Protocol *(*)(SSS_ConfigDict,int,int *))lt_dlsym(hand,"get_state");
  if(!(*get_state) && debug){
    fprintf(stderr,"can't resolve get_state\n");
    return -1;
  }

  *close_conn = (int (*)(struct SSS_Protocol *))lt_dlsym(hand,"close_conn");
  if(!(*close_conn) && debug){
    fprintf(stderr,"can't resolve close_conn\n");
    return -1;
  }

  *serv_connect = (int (*)(struct SSS_Protocol *,int))lt_dlsym(hand,"serv_connect");
  if(!(*serv_connect) && debug){
    fprintf(stderr,"can't resolve serv_connect\n");
    return -1;
  }

  *handle = hand;
  
  return 0;
}

SSS_ServiceList service_cache_init(void){
  SSS_ServiceList ret;
  int i;

  ret = malloc(sizeof(struct SSS_ServiceCache_t));

  for(i=0;i<SSS_MAXPLEX;i++)
    ret->services[i] = NULL;

  return ret;
}

void service_cache_dump(SSS_ServiceList svcs){
  int i;

  for (i=0;i<SSS_MAXPLEX;i++){
    if (svcs->services[i]){
      printf("%s %s %i %s\n",
	     svcs->services[i]->service,
	     svcs->services[i]->host,
	     svcs->services[i]->extra,
	     svcs->services[i]->protocol);
    }
  }
}

void service_cache_add(SSS_ServiceList svcs,struct SSS_Service *service){
  int i;

  for (i=0;(i<SSS_MAXPLEX) && (svcs->services[i]);i++){}

  if (i == SSS_MAXPLEX){
    /* unfortunately, no option at this time but to die - a bug, or SSS_MAXPLEX should be increased! */
    return;
  }

  svcs->services[i] = malloc(sizeof(struct SSS_Service));

  svcs->services[i]->host = my_strdup(service->host);
  svcs->services[i]->extra = service->extra;
  svcs->services[i]->protocol = my_strdup(service->protocol);
  svcs->services[i]->service = my_strdup(service->service);
  svcs->services[i]->schema_version = my_strdup(service->schema_version);
  svcs->services[i]->tier = my_strdup(service->tier);
  svcs->services[i]->valid = 1;

  return;
}

SSS_ServiceList service_cache_lookup(SSS_ServiceList cache,char *service){
  int i;
  SSS_ServiceList ret;

  ret = NULL;

  for (i=0;cache->services[i];i++){
    if (!strcmp(service,cache->services[i]->service)){
      if (cache->services[i]->valid){
	if (!ret){
	  ret = service_cache_init();
	}
	service_cache_add(ret,cache->services[i]);
      }
    }
  }
  
  return ret;
  
}

void service_cache_replace(SSS_ServiceList cache,struct SSS_Service *service){
  int i;

  for (i=0;i<SSS_MAXPLEX;i++){
    if (cache->services[i]){
      if (!strcmp(cache->services[i]->service,service->service)){
	if (cache->services[i]->valid){
	  /* delete this entry because we have a newer one */
	  free(cache->services[i]->service);
	  free(cache->services[i]->host);
	  free(cache->services[i]->protocol);
	  free(cache->services[i]->schema_version);
	  free(cache->services[i]->tier);
	  free(cache->services[i]);
	  cache->services[i] = NULL;
	}
      }
    }
  }

  for (i=0;(i<SSS_MAXPLEX) && (cache->services[i]);i++){}

  if (i == SSS_MAXPLEX){
    /* unfortunately, no option at this time but to die - a bug, or SSS_MAXPLEX should be increased! */
    return;
  }

  cache->services[i] = malloc(sizeof(struct SSS_Service));

  cache->services[i]->host = my_strdup(service->host);
  cache->services[i]->extra = service->extra;
  cache->services[i]->protocol = my_strdup(service->protocol);
  cache->services[i]->service = my_strdup(service->service);
  cache->services[i]->schema_version = my_strdup(service->schema_version);
  cache->services[i]->tier = my_strdup(service->tier);
  cache->services[i]->valid = 1;

  return;
}

void service_cache_free(SSS_ServiceList cache){
  int i;
  for(i=0;i<SSS_MAXPLEX;i++){
    if (cache->services[i]){
      free(cache->services[i]->host);
      free(cache->services[i]->protocol);
      free(cache->services[i]->service);
      free(cache->services[i]->schema_version);
      free(cache->services[i]->tier);
      free(cache->services[i]);
      cache->services[i] = NULL;
    }
  }

  free(cache);
}

void SSS_SetServerProtocol(SSS sss,char *protocol){
  sss->servprotocol = my_strdup(protocol);
}

SSS_ServiceList SSS_StaticLookup(SSS sss,char *service){
  SSS_ServiceList ret;
  ret = service_cache_lookup(sss->service_cache,service);
  if (!ret){
    sss->sss_errno = SSS_SD_LOOKUP_ERROR;
  }
  return ret;
}

int cache_update_from_file(SSS sss,SSS_ServiceList cache,char *conffile){
  struct stat conffile_stat;
  char *file_data;
  int bytes_read;
  int fd;
  XML_Parser parser;
  struct XML_Userdata userdata;

  stat(conffile,&conffile_stat);
  
  file_data = malloc(conffile_stat.st_size+1);

  fd = open(conffile,O_RDONLY);
  if (fd == -1){
    switch(errno){
    case EACCES:
      sss->sss_errno = SSS_PERMISSION_DENIED;
      break;
    case ENAMETOOLONG:
      sss->sss_errno = SSS_NAMETOOLONG;
      break;
    case ENOENT:
      sss->sss_errno = SSS_FILENOTFOUND;
      break;
    }
    free(file_data);
    return -1;
  }

  bytes_read = read(fd,file_data,conffile_stat.st_size);
  while (bytes_read < conffile_stat.st_size)
    bytes_read += read(fd,file_data+bytes_read,conffile_stat.st_size - bytes_read);

  close(fd);

  file_data[conffile_stat.st_size] = '\0';

  parser = XML_ParserCreate(NULL);
    
  userdata.success = 0;
  userdata.messageType = SSS_LOOKUP_RESP;
  userdata.host = 0;
  userdata.extra = 0;
  userdata.protocol = 0;
  userdata.service_cache = cache;

  XML_SetUserData(parser,&userdata);

  XML_SetElementHandler(parser,xml_starthandler,xml_endhandler);

  if (!XML_Parse(parser,file_data,strlen(file_data),1)){
    if (sss->debug){
      fprintf(stderr,"XML Parse error in %s\n",conffile);
      fprintf(stderr,"buf was :%s:\n",file_data);
    }
    XML_ParserFree(parser);
    free(file_data);
    return -1;
  }
  
  return 0;
}

int cache_update_from_sd(SSS sss,char *service){
  struct SSS_Protocol *tempstate;
  int (*prot_connect)(struct SSS_Protocol *,char *,int);
  int (*prot_sconnect)(struct SSS_Protocol *,int);
  int (*prot_close)(struct SSS_Protocol *);
  int (*prot_send)(struct SSS_Protocol *,char *);
  char *(*prot_recv)(struct SSS_Protocol *);
  struct SSS_Protocol * (*prot_getstate)(SSS_ConfigDict,int,int *);
  void *libhandle;
  char *sendbuf,*tempbuf;
  XML_Parser parser;
  struct XML_Userdata userdata;
  int rc;
  int i;
  int error;
  SSS_ConfigDict cfg;

  rc = module_load_funcs(sss->debug,
			 sss->sd_protocol,
			 &prot_connect,
			 &prot_send,
			 &prot_recv,
			 &prot_getstate,
			 &prot_close,
			 &prot_sconnect,
			 &libhandle);

  if (rc == -1){
    sss->sss_errno = SSS_LIBRARY_ERROR;
    return -1;
  }
    
  cfg = NULL;
  for(i=0;i<sss->numconfigs;i++)
    if (!strcmp(sss->configs[i].protocol,sss->sd_protocol))
      cfg = &(sss->configs[i]);
  tempstate = prot_getstate(cfg,sss->debug,&error);
  if (!tempstate){
    sss->sss_errno = error;
    return -1;
  }
  tempstate->debug = sss->debug;
  tempstate->component = my_strdup("service-directory");

  rc = prot_connect(tempstate,sss->sd_host,sss->sd_extra);

  if (rc == -1){
    sss->sss_errno = tempstate->prot_errno;
    prot_close(tempstate);
    lt_dlclose(libhandle);
    return -1;
  }

  sendbuf = my_strdup("<get-location><location component='%s' host='*' port='*' protocol='*' schema_version='*' tier='*'/></get-location>");
  tempbuf = sendbuf;
  /* the +1 is for \0, the -2 is for the format strings */
  sendbuf = malloc(strlen(sss->cluster)+strlen(tempbuf)+strlen(service)+1-2);
  sprintf(sendbuf,tempbuf,service);
  free(tempbuf);

  rc = prot_send(tempstate,sendbuf);

  free(sendbuf);

  if (rc == -1){
    sss->sss_errno = tempstate->prot_errno;
    if(sss->debug)
      fprintf(stderr,"Error in send\n");
    lt_dlclose(libhandle);
    return -1;
  }

  tempbuf = prot_recv(tempstate);

  if (!tempbuf){
    sss->sss_errno = tempstate->prot_errno;
    lt_dlclose(libhandle);
    return -1;
  }

  parser = XML_ParserCreate(NULL);
  
  userdata.success = 0;
  userdata.messageType = SSS_LOOKUP_RESP;
  userdata.host = 0;
  userdata.extra = 0;
  userdata.protocol = 0;
  userdata.overwrite = 1;
  userdata.service_cache = sss->service_cache;

  XML_SetUserData(parser,&userdata);

  XML_SetElementHandler(parser,xml_starthandler,xml_endhandler);

  if (!XML_Parse(parser,tempbuf,strlen(tempbuf),1)){
    if (sss->debug)
      fprintf(stderr,"XML parse error, buffer was :%s:\n",tempbuf);
    sss->sss_errno = SSS_XML_PARSE;
    XML_ParserFree(parser);
    free(tempbuf);
    lt_dlclose(libhandle);
    return -1;
  }

  free(tempbuf);

  XML_ParserFree(parser);

  if (!userdata.success){
    if (sss->debug)
      fprintf(stderr,"Service directory returned failure: no such component\n");
    /* should have more errno granularity */
    sss->sss_errno = SSS_SD_LOOKUP_ERROR;

    prot_close(tempstate);
    lt_dlclose(libhandle);
    return -1;
  }

  prot_close(tempstate);

  lt_dlclose(libhandle);

  if (!userdata.host){
    if (sss->debug)
      fprintf(stderr,"Service directory did not specify a host!\n");
    sss->sss_errno = SSS_SD_FAIL;
    lt_dlclose(libhandle);
    return -1;
  }
  if (!userdata.extra){
    if (sss->debug)
      fprintf(stderr,"Service directory did not specify a port!\n");
    sss->sss_errno = SSS_SD_FAIL;
    lt_dlclose(libhandle);
    return -1;
  }
  if (!userdata.protocol){
    if (sss->debug)
      fprintf(stderr,"Service directory did not specify a protocol!\n");
    sss->sss_errno = SSS_SD_FAIL;
    lt_dlclose(libhandle);
    return -1;
  }

  return 0;
  
}

SSS_ServiceList find_from_cache(SSS sss,char *service,char *cachefile){
  SSS_ServiceList ret;
  struct stat conffile_stat;

  if (sss->debug)
    printf("find_from_cache: looking for :%s:\n",service);

  ret = service_cache_lookup(sss->service_cache,service);

  if (!ret){
    if (sss->debug)
      printf("not in the memory cache\n");

    /* not in the memory cache */

    stat(cachefile,&conffile_stat);

#ifdef __APPLE__
    if (conffile_stat.st_mtimespec.tv_sec > sss->sss_conf_mtime.tv_sec){
      sss->sss_conf_mtime = conffile_stat.st_mtimespec;
#else
    if (conffile_stat.st_mtime > sss->sss_conf_mtime){
      sss->sss_conf_mtime = conffile_stat.st_mtime;
#endif

      /* file has been updated */
      if (sss->debug)
	printf("file has been updated\n");
      cache_update_from_file(sss,sss->service_cache,cachefile);
      
      return find_from_cache(sss,service,cachefile);
    }
    else{
      if (sss->debug)
	printf("file has NOT been updated, going to sd\n");
      /* file not updated, go to service directory */
      if (sss->sd_host){
        cache_update_from_sd(sss,service);
        ret = service_cache_lookup(sss->service_cache,service);
        if (!ret){
          if(sss->debug)
            printf("was not in the sd\n");
          /* wasn't in the sd, now we can fail */
          sss->sss_errno = SSS_SD_LOOKUP_ERROR;
          return NULL;
        }
        else{
          if(sss->debug)
            printf("was in the sd\n");
          /* the sd must have had it */
          return ret;
        }
      }
      else{
        sss->sss_errno = SSS_SD_LOOKUP_ERROR;
        return NULL;
      }
    }
  }
  else{
    if(sss->debug)
      printf("found in the memory cache\n");
    /* in the memory cache */
    return ret;
  }
}

int cached_connect(SSS sss,SSS_Connection conn){
  int rc;
  int i;
  int conffile_name_len;
  char *conffile_name;
  struct stat conffile_stat;
  char *component;
  int error;
  SSS_ServiceList svcs;
  SSS_ConfigDict cfg;

  if(sss->debug)
    printf("cached_connect: starting on :%s:\n",conn->state->component);

  rc = conn->cli_connect(conn->state,conn->host,conn->extra);

  if (rc == -1){
    if(sss->debug)
      printf("the cached data failed us!\n");
    /* The cached data is incorrect! */

    /* first check the file */

    conffile_name_len = snprintf(buf,1,"%s/sss.conf",PATH_TO_CONFIG);

    conffile_name = malloc(conffile_name_len+1);

    sprintf(conffile_name,"%s/sss.conf",PATH_TO_CONFIG);

    stat(conffile_name,&conffile_stat);

#ifdef __APPLE__
    if (conffile_stat.st_mtimespec.tv_sec > sss->sss_conf_mtime.tv_sec){
#else
    if (conffile_stat.st_mtime > sss->sss_conf_mtime){
#endif
      if(sss->debug)
	printf("file has been updated\n");
      /* configuration file has been updated */

      rc = cache_update_from_file(sss,sss->service_cache,conffile_name);
      if (rc == -1)
	return -1;

      /* try again */

      return cached_connect(sss,conn);
    }
    else{

      free(conffile_name);
      
      if(sss->debug)
	printf("file has NOT been updated\n");

      /* configuration file not updated - check the sd! */

      if (sss->sd_host){
        rc = cache_update_from_sd(sss,conn->state->component);

        if (rc == -1){
          if(sss->debug)
            printf("service directory failed in cache_update_from_sd!\n");
          /* service directory failure */
          return -1;
        }

        svcs = service_cache_lookup(sss->service_cache,conn->state->component);

        if (!svcs){
          if(sss->debug)
            printf("nothing from the sd\n");
          /* nothing from the sd, so fail! */
          return -1;
        }
      }
      else{
        if(sss->debug)
          printf("no service directory\n");
        return -1;
      }

      for (i=0;svcs->services[i];i++){

	if (strcmp(svcs->services[i]->protocol,
		   conn->protocol)){
	  /* the protocol we had was wrong.  We can no longer just
	     reconnect - a new module must be loaded.  */

	  if (sss->debug)
	    printf("Old protocol was %s, new protocol is %s\n",conn->protocol,sss->service_cache->services[i]->protocol);

	  if(sss->debug)
	    printf("Wrong protocol! reloading module...\n");

	  free(conn->host);
	  free(conn->protocol);
	  conn->host = my_strdup(svcs->services[i]->host);
	  conn->extra = svcs->services[i]->extra;
	  conn->protocol = my_strdup(svcs->services[i]->protocol);
	  component = my_strdup(conn->state->component);
	  conn->close(conn->state);
	  rc = module_load_funcs(sss->debug,
				 conn->protocol,
				 &conn->cli_connect,
				 &conn->send,
				 &conn->recv,
				 &conn->getstate,
				 &conn->close,
				 &conn->serv_connect,
				 &conn->libhandle);
	
	  if (rc == -1){
	    sss->sss_errno = SSS_LIBRARY_ERROR;
	    free(conn->handle);
	    free(conn->host);
	    free(conn->protocol);
	    free(conn);
	    service_cache_free(svcs);
	    return -1;
	  }

	  cfg = NULL;
	  for(i=0;i<sss->numconfigs;i++){
	    if (!strcmp(sss->configs[i].protocol,conn->protocol))
	      cfg = &(sss->configs[i]);
	  }
	  
	  conn->state = conn->getstate(cfg,sss->debug,&error);
	  if (!conn->state){
	    sss->sss_errno = error;
	    free(conn->handle);
	    free(conn->host);
	    free(conn->protocol);
	    free(conn);
	    service_cache_free(svcs);
	    return -1;
	  }
	  conn->state->debug = sss->debug;
	  conn->state->component = component;

	}

	/* try again, with the new values */

	rc = conn->cli_connect(conn->state,
			       svcs->services[i]->host,
			       svcs->services[i]->extra);

	if(!rc){
	  if(sss->debug)
	    printf("success! updating conn data...\n");
	  /* success! so update the connection with the real values */
	  free(conn->host);
	  free(conn->protocol);
	  conn->host = my_strdup(svcs->services[i]->host);
	  conn->extra = svcs->services[i]->extra;
	  conn->protocol = my_strdup(svcs->services[i]->protocol);
	  service_cache_free(svcs);
	  return rc;
	}
      }
      service_cache_free(svcs);
    }
  }
  return rc;
}

SSS SSS_Init(char *conffile,int debug){
  int fd,rc;
  SSS sss;
  char *file;
  struct stat retstat;
  struct stat sss_conf_stat;
  char linebuf[1024];
  char charbuf[1];
  ssize_t bytes_read;
  int pos;
  int allGood;
  char *keybuf;
  char *valbuf;
  char *myconf;
  char confprotocol[128];
  int i,j;
  char *temp;
  int numprefs;
  int filename_len;
  XML_Parser parser;
  struct XML_Userdata userdata;
  char *sss_conf_buf;
  FILE *filep;
  SSS_ServiceList sds;
  
  lt_dlinit();

  sss = (SSS)malloc(sizeof(struct SSS_t));

  for(i=0;i<SSS_MAXPLEX;i++)
    sss->connections[i] = NULL;
  
  sss->debug = debug;
  sss->servprotocol = NULL;
  sss->numconfigs = 0;
  sss->configs = NULL;
  sss->protoprefs = malloc(sizeof(char *));
  sss->cluster = NULL;
  sss->msglog = NULL;
  sss->connection_index = 0;
  sss->service_cache = NULL;
  confprotocol[0] = '\0';
  
  /* read the service cache file */

  filename_len = snprintf(buf,1,"%s/sss.conf",PATH_TO_CONFIG);

  myconf = malloc(filename_len+100);

  sprintf(myconf,"%s/sss.conf",PATH_TO_CONFIG);

  rc = stat(myconf,&sss_conf_stat);

  if (rc == -1){
    free(myconf);
    free(sss->protoprefs);
    free(sss);
    return NULL;
  }

  fd = open(myconf,O_RDONLY);
  if (fd == -1){
    free(myconf);
    free(sss->protoprefs);
    free(sss);
    return NULL;
  }

  sss_conf_buf = malloc(sss_conf_stat.st_size+1);

#ifdef __APPLE__
  sss->sss_conf_mtime = sss_conf_stat.st_mtimespec;
#else
  sss->sss_conf_mtime = sss_conf_stat.st_mtime;
#endif

  bytes_read = read(fd,sss_conf_buf,sss_conf_stat.st_size);

  while (bytes_read < sss_conf_stat.st_size){
    bytes_read += read(fd,sss_conf_buf+bytes_read,sss_conf_stat.st_size - bytes_read);
  }

  sss_conf_buf[sss_conf_stat.st_size] = '\0';

  close(fd);

  free(myconf);

  /* get service directory information from file */

  parser = XML_ParserCreate(NULL);

  userdata.success = 0;
  userdata.messageType = SSS_LOOKUP_RESP;
  userdata.overwrite = 0;

  userdata.service_cache = service_cache_init();

  XML_SetUserData(parser,&userdata);

  XML_SetElementHandler(parser,xml_starthandler,xml_endhandler);

  if (!XML_Parse(parser,sss_conf_buf,strlen(sss_conf_buf),1)){
    if (sss->debug)
      fprintf(stderr,"XML parse error in sss.conf file.\n");
    service_cache_free(userdata.service_cache);
    return NULL;
  }

  free(sss_conf_buf);

  sss->service_cache = userdata.service_cache;

  sds = service_cache_lookup(sss->service_cache,"service-directory");

  if (sds){
    sss->sd_host = my_strdup(sds->services[0]->host);
    sss->sd_extra = sds->services[0]->extra;
    sss->sd_protocol = my_strdup(sds->services[0]->protocol);
    sss->sd_schema_version = my_strdup(sds->services[0]->schema_version);
    sss->sd_tier = my_strdup(sds->services[0]->tier);
  }
  else{
    sss->sd_host = NULL;
    sss->sd_extra = 0;
    sss->sd_protocol = NULL;
    sss->sd_schema_version = NULL;
    sss->sd_tier = NULL;
  }

  if (sds)
    service_cache_free(sds);

  /* if file is null go for defaults */
  if (!conffile){
    temp = getenv("HOME");
    rc = 0;
    if (temp){
      myconf=malloc(strlen(temp)+strlen("/.sssrc")+1);
      sprintf(myconf,"%s/.sssrc",temp);
      rc = stat(myconf,&retstat);
    }
    if (rc == -1 || !temp) {
      /* use default file */
      file = malloc(strlen(PATH_TO_CONFIG)+strlen("/ssslib.conf")+2);
      sprintf(file, "%s/ssslib.conf", PATH_TO_CONFIG);
    } else {
      file = my_strdup(myconf);
      free(myconf);
    }
  }
  else
    file = my_strdup(conffile);

  /* check for conf file existence */
  
  rc = stat(file,&retstat);
  if (rc == -1){
    if (debug)
      fprintf(stderr,"Error getting file information for file :%s:! Maybe you specified an incorrect path? Check errno!\n",file);
    free(file);
    service_cache_free(sss->service_cache);
    return NULL;
  }

  filep = fopen(file,"r");
  if (!filep){
    if (debug)
      fprintf(stderr,"Error opening file! Check errno!\n");
    free(file);
    service_cache_free(sss->service_cache);
    return NULL;
  }

  bytes_read = fread(charbuf,1,1,filep);
  if (bytes_read == 0 && ferror(filep)){
    if (debug)
      fprintf(stderr,"Read error! Check errno!\n");
    free(file);
    fclose(filep);
    service_cache_free(sss->service_cache);
    return NULL;
  }
  pos = 0;
  allGood = 1;

  while ((bytes_read && !feof(filep)) && allGood){
    linebuf[pos] = charbuf[0];
    if (linebuf[pos] == '\n'){
      linebuf[pos] = '\0';
      pos = 0;
      /* parse key-val pair */
      /* work some magic */
      /* first, check to see if we are in a protocol-specific config section */
      if (sscanf(linebuf,"[%s",confprotocol)>0){
	/* now we are in a specific config part */
	/* so increase the number of configs and prepare to enter keyval pairs */

	confprotocol[strlen(confprotocol)-1] = '\0';

	sss->numconfigs++;
	if (sss->configs)
	  sss->configs = realloc(sss->configs,sss->numconfigs * sizeof(struct SSS_ConfigDict_t));
	else
	  sss->configs = malloc(sizeof(struct SSS_ConfigDict_t));

	strncpy(sss->configs[sss->numconfigs-1].protocol,confprotocol,128);

	sss->configs[sss->numconfigs-1].keywords = malloc(sizeof(char *));
	sss->configs[sss->numconfigs-1].values = malloc(sizeof(char *));

	sss->configs[sss->numconfigs-1].keywords[0] = NULL;
	sss->configs[sss->numconfigs-1].values[0] = NULL;

      }
	
      valbuf = strstr(linebuf,"=");
      if (valbuf){
	valbuf++;
	keybuf = linebuf;
	keybuf[valbuf-keybuf-1] = '\0';
	if (confprotocol[0] != '\0'){
	  for(i=0;i<sss->numconfigs;i++){
	    if (!strcmp(sss->configs[i].protocol,confprotocol)){
	      for(j=0;sss->configs[i].keywords[j];j++){}
	      sss->configs[i].keywords = realloc(sss->configs[i].keywords,(j+2)*sizeof(char *));
	      sss->configs[i].values = realloc(sss->configs[i].values,(j+2)*sizeof(char *));
	      sss->configs[i].keywords[j+1] = NULL;
	      sss->configs[i].values[j+1] = NULL;
	      sss->configs[i].values[j] = my_strdup(valbuf);
	      sss->configs[i].keywords[j] = my_strdup(keybuf);
	    }
	  }
	}
	else if (!strcmp(keybuf,"cluster"))
	  sss->cluster = my_strdup(valbuf);
	else if (!strcmp(keybuf,"msg_log"))
	  sss->msglog = my_strdup(valbuf);
	else if (!strcmp(keybuf,"prefs")){
	  numprefs = 0;
	  while ((temp = strstr(valbuf,","))){
	    valbuf[temp - valbuf] = '\0';
	    temp++;
	    numprefs++;
	    sss->protoprefs = realloc(sss->protoprefs,(numprefs+1)*sizeof(char *));
	    sss->protoprefs[numprefs] = NULL;
	    sss->protoprefs[numprefs - 1] = my_strdup(valbuf);
	    valbuf = temp;
	  }
	  numprefs++;
	  sss->protoprefs = realloc(sss->protoprefs,(numprefs+1)*sizeof(char *));
	  sss->protoprefs[numprefs] = NULL;
	  sss->protoprefs[numprefs - 1] = my_strdup(valbuf);
	}
      }
    }
    else
      pos++;
    if (pos > 1022){
      /* should make this dynamic */
      if (debug)
	fprintf(stderr,"Line in config file too long! aborting read...\n");
      allGood = 0;
    }
    bytes_read = fread(charbuf,1,1,filep);
    if (bytes_read == 0 && ferror(filep)){
      if (debug)
	fprintf(stderr,"Read error! Check errno! Aborting config file read...\n");
      allGood = 0;
      free(file);
      fclose(filep);
      service_cache_free(sss->service_cache);
      return NULL;
    }
  }

  free(file);

  close(fd);
  
  return sss;
  
}

int SSS_SDRegister(SSS sss,char *service,char *host,int port,char *protocol,char *schema_version,char *tier){
  int rc,i;
  void *libhandle;
  struct SSS_Protocol *tempstate;
  int (*prot_connect)(struct SSS_Protocol *,char *,int);
  int (*prot_sconnect)(struct SSS_Protocol *,int);
  int (*prot_close)(struct SSS_Protocol *);
  int (*prot_send)(struct SSS_Protocol *,char *);
  char *(*prot_recv)(struct SSS_Protocol *);
  struct SSS_Protocol * (*prot_getstate)(SSS_ConfigDict,int,int *);
  char *sendbuf,*tempbuf;
  XML_Parser parser;
  struct XML_Userdata userdata;
  int error;
  SSS_ConfigDict cfg;

  if (sss->debug)
    printf("ssslib.c: SDRegister starting\n");

  sss->servprotocol = my_strdup(protocol);
  
  rc = module_load_funcs(sss->debug,
			 sss->sd_protocol,
			 &prot_connect,
			 &prot_send,
			 &prot_recv,
			 &prot_getstate,
			 &prot_close,
			 &prot_sconnect,
			 &libhandle);

  if (rc == -1){
    sss->sss_errno = SSS_LIBRARY_ERROR;
    return -1;
  }
    
  
  cfg = NULL;
  for(i=0;i<sss->numconfigs;i++)
    if (!strcmp(sss->configs[i].protocol,sss->sd_protocol))
      cfg = &(sss->configs[i]);

  tempstate = prot_getstate(cfg,sss->debug,&error);
  if (!tempstate){
    sss->sss_errno = error;
    return -1;
  }
  tempstate->debug = sss->debug;
  tempstate->component = my_strdup("service-directory");

  rc = prot_connect(tempstate,sss->sd_host,sss->sd_extra);

  if (rc == -1){
    if (sss->debug)
      fprintf(stderr,"ssslib.c: SDRegister: prot_connect returned -1\n");
    sss->sss_errno = tempstate->prot_errno;
    
    lt_dlclose(libhandle);
    return -1;
  }

  if (!schema_version)
    schema_version = "1.0";
  if (!tier)
    tier = "1";

  tempbuf = my_strdup("<add-location><location component='%s' host='%s' port='%i' protocol='%s' schema_version='%s' tier='%s'/></add-location>");
  sendbuf = malloc(1+snprintf(buf,1,tempbuf,service,host,port,protocol,schema_version,tier));

  sprintf(sendbuf,tempbuf,service,host,port,protocol,schema_version,tier);
  free(tempbuf);

  rc = prot_send(tempstate,sendbuf);
	
  free(sendbuf);

  if (rc == -1){
    sss->sss_errno = tempstate->prot_errno;
    if(sss->debug)
      fprintf(stderr,"ssslib.c: SDRegister: Error in send\n");
    
    lt_dlclose(libhandle);
    return -1;
  }

  tempbuf = prot_recv(tempstate);

  if (!tempbuf){
    sss->sss_errno = tempstate->prot_errno;
    if(sss->debug){
      fprintf(stderr,"ssslib.c: SDRegister: Error in receive!\n");
    }
    lt_dlclose(libhandle);
    return -1;
  }

  parser = XML_ParserCreate(NULL);
  
  userdata.success = 0;
  userdata.messageType = SSS_REG_RESP;
  userdata.overwrite = 0;

  XML_SetUserData(parser,&userdata);

  XML_SetElementHandler(parser,xml_starthandler,xml_endhandler);

  if (!XML_Parse(parser,tempbuf,strlen(tempbuf),1)){
    if (sss->debug)
      fprintf(stderr,"XML parse error, buffer was :%s:\n",tempbuf);
    sss->sss_errno = SSS_XML_PARSE;
    free(tempbuf);
    XML_ParserFree(parser);
    
    lt_dlclose(libhandle);
    return -1;
  }

  free(tempbuf);

  XML_ParserFree(parser);
  
  if (!userdata.success){
    if (sss->debug)
      fprintf(stderr,"Service directory returned failure: %s\n",userdata.reason);
    /* should have more errno granularity */
    sss->sss_errno = SSS_SD_FAIL;
    prot_close(tempstate);
    lt_dlclose(libhandle);
    return -1;
  }

  prot_close(tempstate);
    
  lt_dlclose(libhandle);

  if (sss->debug)
    printf("ssslib.c: SDRegister returning success\n");

  return 0;
  
}


int SSS_SDRemove(SSS sss,char *service,char *host,int port,char *protocol){
  int rc,i;
  void *libhandle;
  struct SSS_Protocol *tempstate;
  int (*prot_connect)(struct SSS_Protocol *,char *,int);
  int (*prot_sconnect)(struct SSS_Protocol *,int);
  int (*prot_close)(struct SSS_Protocol *);
  int (*prot_send)(struct SSS_Protocol *,char *);
  char *(*prot_recv)(struct SSS_Protocol *);
  struct SSS_Protocol * (*prot_getstate)(SSS_ConfigDict,int,int *);
  char *sendbuf,*tempbuf;
  XML_Parser parser;
  struct XML_Userdata userdata;
  int error;
  SSS_ConfigDict cfg;

  if (sss->debug)
    printf("ssslib.c: SDRemove starting\n");

  rc = module_load_funcs(sss->debug,
			 sss->sd_protocol,
			 &prot_connect,
			 &prot_send,
			 &prot_recv,
			 &prot_getstate,
			 &prot_close,
			 &prot_sconnect,
			 &libhandle);

  if (rc == -1){
    sss->sss_errno = SSS_LIBRARY_ERROR;
    return -1;
  }
    
  cfg = NULL;

  for(i=0;i<sss->numconfigs;i++)
    if (!strcmp(sss->configs[i].protocol,sss->sd_protocol))
      cfg = &(sss->configs[i]);

  tempstate = prot_getstate(cfg,sss->debug,&error);
  if (!tempstate){
    sss->sss_errno = error;
    return -1;
  }
  tempstate->debug = sss->debug;
  tempstate->component = my_strdup("service-directory");

  rc = prot_connect(tempstate,sss->sd_host,sss->sd_extra);

  if (rc == -1){
    if (sss->debug)
      fprintf(stderr,"ssslib.c: SDRemove: prot_connect returned -1\n");
    sss->sss_errno = tempstate->prot_errno;
    lt_dlclose(libhandle);
    return -1;
  }

  tempbuf = my_strdup("<del-location><location component='%s' host='%s' port='%d' protocol='%s'/></del-location>");
  sendbuf = malloc(1+snprintf(buf,1,tempbuf,service,host,port,protocol));

  sprintf(sendbuf,tempbuf,service,host,port,protocol);

  free(tempbuf);

  rc = prot_send(tempstate,sendbuf);

  if (rc == -1){
    sss->sss_errno = tempstate->prot_errno;
    if(sss->debug)
      fprintf(stderr,"Error in send\n");
    lt_dlclose(libhandle);
    return -1;
  }

  tempbuf = prot_recv(tempstate);

  if (!tempbuf){
    sss->sss_errno = tempstate->prot_errno;
    if(sss->debug)
      fprintf(stderr,"Error in receive!\n");
    lt_dlclose(libhandle);
    return -1;
  }

  parser = XML_ParserCreate(NULL);
  
  userdata.success = 0;
  userdata.messageType = SSS_REG_RESP;

  XML_SetUserData(parser,&userdata);

  XML_SetElementHandler(parser,xml_starthandler,xml_endhandler);

  if (!XML_Parse(parser,tempbuf,strlen(tempbuf),1)){
    if (sss->debug)
      fprintf(stderr,"XML parse error, buffer was :%s:\n",tempbuf);
    sss->sss_errno = SSS_XML_PARSE;
    free(tempbuf);
    lt_dlclose(libhandle);
    return -1;
  }

  free(tempbuf);

  XML_ParserFree(parser);
  
  if (!userdata.success){
    if (sss->debug)
      fprintf(stderr,"Service directory returned failure: %s\n",userdata.reason);
    /* should have more errno granularity */
    sss->sss_errno = SSS_SD_FAIL;
    prot_close(tempstate);
    lt_dlclose(libhandle);
    return -1;
  }

  prot_close(tempstate);

  lt_dlclose(libhandle);

  if (sss->debug)
    printf("ssslib.c: SDRemove returning success\n");

  return 0;
  
}

SSS_ServiceList SSS_SDLookup(SSS sss,char *service){
  int conffile_name_len;
  char *conffile_name;
  SSS_ServiceList svcs;

  if (sss->debug)
    printf("ssslib.c: SDLookup starting\n");
    
  /* check the cache first */

  conffile_name_len = snprintf(buf,1,"%s/sss.conf",PATH_TO_CONFIG);
  conffile_name = malloc(conffile_name_len+1);
  sprintf(conffile_name,"%s/sss.conf",PATH_TO_CONFIG);

  svcs = find_from_cache(sss,service,conffile_name);

  if (!svcs){
    /* couldn't find it, anywhere, or some error with the sd */
    /* assume the find_from_cache set sss->sss_errno */
    free(conffile_name);
    if (sss->debug)
      printf("ssslib.c: SDLookup returning failure\n");
    return NULL;
  }

  free(conffile_name);

  if (sss->debug)
    printf("ssslib.c: SDLookup returning success\n");
    
  return svcs;

}

char *SSS_ServerInit(SSS sss,int extra){

  SSS_Connection conn;
  int rc;
  struct sockaddr_in peer;
#ifndef __APPLE__
  socklen_t size;
#else
  int size;
#endif
  int i;
  int error;
  SSS_ConfigDict cfg;
  int sock = extra;
  
  if (sss->debug)
    printf("ssslib.c: ServerInit starting\n");

  conn = malloc(sizeof(struct SSS_Connection_t));

  conn->libhandle = NULL;
  conn->handle = NULL;
  conn->protocol = NULL;
  conn->host = NULL;
  conn->extra = -1;

  if (sss->servprotocol){
    if (sss->debug)
      printf("ssslib.c: ServerInit: using servprotocol, which is :%s:\n",sss->servprotocol);
    conn->protocol = my_strdup(sss->servprotocol);
  }
  else{
    if (sss->debug)
      printf("ssslib.c: ServerInit: Never SDRegistered, assuming we are Service Directory\n");
    conn->protocol = my_strdup(sss->sd_protocol);
  }

  rc = getpeername(sock,(struct sockaddr *)&peer,&size);

  if (rc == -1){
    switch (errno){
    case EBADF:
      sss->sss_errno = SSS_BAD_FD;
      break;
    case ENOTSOCK:
      sss->sss_errno = SSS_NOT_A_SOCKET;
      break;
    case ENOTCONN:
      sss->sss_errno = SSS_SOCK_NOT_CONNECTED;
      break;
    case ENOBUFS:
      sss->sss_errno = SSS_NOMEM;
      break;
    case EFAULT:
      sss->sss_errno = SSS_FAULT;
      break;
    case EINVAL:
      sss->sss_errno = SSS_INVAL_ARG;
      break;
    }
    return NULL;
  }
  
  conn->handle = malloc(11+1+1+15+32);

  sprintf(conn->handle,"%s:%i:%d",inet_ntoa(peer.sin_addr),ntohs(peer.sin_port), 
	  ++sss->connection_index);

  if (sss->debug)
    printf("ssslib.c: ServerInit: handle for connections is %s\n",conn->handle);

  if (strlen(conn->protocol) > 124){
    if (sss->debug)
      fprintf(stderr,"Protocol name too long!\n");
    /* this should be deprecated */
    sss->sss_errno = SSS_PROT_NAME_LONG;
    free(conn->handle);
    free(conn);
    return NULL;
  }

  rc = module_load_funcs(sss->debug,
			 conn->protocol,
			 &conn->cli_connect,
			 &conn->send,
			 &conn->recv,
			 &conn->getstate,
			 &conn->close,
			 &conn->serv_connect,
			 &conn->libhandle);

  if (rc == -1){
    sss->sss_errno = SSS_LIBRARY_ERROR;
    free(conn->handle);
    free(conn);
    return NULL;
  }
    
  cfg = NULL;

  for(i=0;i<sss->numconfigs;i++)
    if (!strcmp(sss->configs[i].protocol,conn->protocol))
      cfg = &(sss->configs[i]);

  conn->state = conn->getstate(cfg,sss->debug,&error);
  if (!conn->state){
    sss->sss_errno = error;
    free(conn->handle);
    free(conn);
    return NULL;
  }

  conn->state->debug = sss->debug;

  conn->state->component = NULL;

  rc = conn->serv_connect(conn->state,extra);

  if (rc){
    sss->sss_errno = conn->state->prot_errno;
    lt_dlclose(conn->libhandle);
    free(conn->handle);
    free(conn);
    return NULL;
  }

  for (i=0;i<SSS_MAXPLEX;i++)
    if (!sss->connections[i]){
      if (sss->debug)
	printf("found place for connection in array slot %i\n",i);
      sss->connections[i] = conn;
      break;
    }

  if (sss->debug)
    printf("ssslib.c: ServerInit returning successfully\n");
  
  return conn->handle;
  
}

char *SSS_ClientInit(SSS sss,char *service){
  SSS_ServiceList svcs;
  SSS_Connection conn;
  int rc,i;
  char *tempprot;
  int error;
  SSS_ConfigDict cfg;

  if (sss->debug)
    printf("ssslib.c: ClientInit starting\n");

  svcs = SSS_SDLookup(sss,service);
  if (!svcs)
    return NULL;

  conn = malloc(sizeof(struct SSS_Connection_t));

  conn->host = my_strdup(svcs->services[0]->host);
  conn->extra = svcs->services[0]->extra;
  conn->protocol = my_strdup(svcs->services[0]->protocol);

  conn->handle = malloc(32*sizeof(char)); 
  sprintf(conn->handle,"%i",++sss->connection_index);

  service_cache_free(svcs);

  if (strlen(conn->protocol) > 124){
    if (sss->debug)
      fprintf(stderr,"Protocol name too long!\n");
    sss->sss_errno = SSS_PROT_NAME_LONG;
    free(conn->host);
    free(conn->protocol);
    free(conn->handle);
    free(conn);
    return NULL;
  }

  rc = module_load_funcs(sss->debug,
			 conn->protocol,
			 &conn->cli_connect,
			 &conn->send,
			 &conn->recv,
			 &conn->getstate,
			 &conn->close,
			 &conn->serv_connect,
			 &conn->libhandle);

  if (rc == -1){
    sss->sss_errno = SSS_LIBRARY_ERROR;
    free(conn->handle);
    free(conn->host);
    free(conn->protocol);
    free(conn);
    return NULL;
  }
    
  cfg = NULL;

  for(i=0;i<sss->numconfigs;i++)
    if (!strcmp(sss->configs[i].protocol,conn->protocol))
      cfg = &(sss->configs[i]);

  conn->state = conn->getstate(cfg,sss->debug,&error);
  if (!conn->state){
    sss->sss_errno = error;
    free(conn->handle);
    free(conn->host);
    free(conn->protocol);
    free(conn);
    return NULL;
  }

  conn->state->debug = sss->debug;

  conn->state->component = my_strdup(service);
  
  tempprot = my_strdup(conn->protocol);

  rc = cached_connect(sss,conn);

  if (rc == -1){
    sss->sss_errno = conn->state->prot_errno;
    conn->close(conn->state);
    free(conn->host);
    free(conn->protocol);
    free(conn->handle);
    free(tempprot);
    lt_dlclose(conn->libhandle);
    free(conn);
    return NULL;
  }

  free(tempprot);

  for(i=0;i<SSS_MAXPLEX;i++)
    if (!sss->connections[i]){
      if (sss->debug)
	printf("place for connection found in array slot %i\n",i);
      sss->connections[i] = conn;
      break;
    }

  if (sss->debug)
    printf("ssslib.c: ClientInit returning handle :%s:\n",conn->handle);

  return conn->handle;
  
}

int SSS_ClientClose(SSS sss,char *handle){
  int i,done;
  int rc;

  if (sss->debug)
    printf("ssslib.c: ClientClose starting\n");

  done = 0;
  
  for (i=0;i<SSS_MAXPLEX;i++){
    if (sss->connections[i]){
      if (!strcmp(sss->connections[i]->handle,handle)){
	free(sss->connections[i]->handle);
	if (sss->connections[i]->host)
	  free(sss->connections[i]->host);
	sss->connections[i]->close(sss->connections[i]->state);
	rc = lt_dlclose(sss->connections[i]->libhandle);
	free(sss->connections[i]->protocol);
	free(sss->connections[i]);
	sss->connections[i] = NULL;
	done = 1;
      }
    }
  }

  if (!done){
    if (sss->debug)
      fprintf(stderr,"Can't find connection for handle %s!\n",handle);
    sss->sss_errno = SSS_NO_SUCH_CONN;
    return -1;
  }
  
  if (sss->debug)
    printf("ssslib.c: ClientClose returning successfully\n");

  return 0;
  
}

int SSS_ServerClose(SSS sss,char *handle){
  int i,done;

  if (sss->debug)
    printf("ssslib.c: ServerClose starting with handle %s\n",handle);

  done = 0;
  for (i=0;i<SSS_MAXPLEX;i++){
    if (sss->connections[i]){
      if (!strcmp(sss->connections[i]->handle,handle)){
	free(sss->connections[i]->handle);
	sss->connections[i]->close(sss->connections[i]->state);
	lt_dlclose(sss->connections[i]->libhandle);
	free(sss->connections[i]->protocol);
	free(sss->connections[i]);
	sss->connections[i] = NULL;
	done = 1;
      }
    }
  }

  if (!done){
    if (sss->debug)
      fprintf(stderr,"Can't find connection for handle %s!\n",handle);
    sss->sss_errno = SSS_NO_SUCH_CONN;
    return -1;
  }

  if (sss->debug)
    printf("ssslib.c: ServerClose returning successfully\n");

  return 0;
}

int SSS_SendMessage(SSS sss,char *handle,char *message){
  int i,rc,sent,fd,logmsg_size,curtime;

  struct SSS_Protocol *state;
  
  struct flock fl;

  char *logmsg;

  sent = 0;

  if (sss->debug)
    printf("ssslib.c: SendMessage starting with msg=:%s:\n",message);

  for (i=0;i<SSS_MAXPLEX;i++){
    if (sss->connections[i]){
      if (!strcmp(sss->connections[i]->handle,handle)){
	state = sss->connections[i]->state;
	rc = sss->connections[i]->send(state,message);
	if (rc == -1){
	  if (sss->debug)
	    fprintf(stderr,"error in send!\n");
	  sss->sss_errno = state->prot_errno;
	  return -1;
	}
	sent = 1;
	if (sss->msglog){
	  fl.l_type = F_WRLCK;
	  fl.l_whence = SEEK_SET;
	  fl.l_start = 0;
	  fl.l_len = 0;
	  fl.l_pid = getpid();
	  fd = open(sss->msglog,O_WRONLY | O_CREAT | O_APPEND);
	  chmod(sss->msglog,0644);
	  if (fd == -1)
	    perror("Warning: couldn't open msg log file: ");
	  else{
	    rc = fcntl(fd,F_SETLKW,&fl);
	    if (rc == -1)
	      perror("Warning: error locking msg log file: ");
	    else{
	      if (state->component){
		curtime = time(NULL);
		logmsg_size = snprintf(buf,1,"%i:%s:%s\n\n",curtime,message,state->component) + 1;
		logmsg = malloc(logmsg_size);
		sprintf(logmsg,"%i:%s:%s\n\n",curtime,message,state->component);
		rc = write(fd,logmsg,logmsg_size-1);
		free(logmsg);
	      }
	      fl.l_type = F_UNLCK;
	      rc = fcntl(fd,F_SETLKW,&fl);
	      rc = close(fd);
	    }
	  }
	}
      }
    }
  }

  if (!sent && sss->debug)
    printf("ssslib.c: SendMessage: it was never sent!\n");

  if (sss->debug)
    printf("ssslib.c: SendMessage returning success\n");

  return 0;
  
}

char *SSS_RecvMessage(SSS sss,char *handle){
  int i;

  char *retbuf = NULL;

  if (sss->debug)
    printf("ssslib.c: RecvMessage starting\n");

  for (i=0;i<SSS_MAXPLEX;i++){
    if (sss->connections[i]){
      if (!strcmp(sss->connections[i]->handle,handle)){
	retbuf = sss->connections[i]->recv(sss->connections[i]->state);
	if (!retbuf){
	  if (sss->connections[i]->state->prot_errno == SSS_EOF){
	    sss->sss_errno = SSS_EOF;
	  }
	  else
	    sss->sss_errno = sss->connections[i]->state->prot_errno;
	  return NULL;
	}
      }
    }
  }

  if (sss->debug)
    printf("ssslib.c: RecvMessage returning success and msg=:%s:\n",retbuf);

  return retbuf;
  
}

int SSS_Close(SSS sss){
  int i,j;
  int tempdebug = sss->debug;

  if (tempdebug)
    printf("ssslib.c: SSS_Close starting\n");
  
  if (sss->debug)
    for (i=0;i<SSS_MAXPLEX;i++)
      if (sss->connections[i])
	if (sss->debug)
	  fprintf(stderr, "error - there are connections that have not been properly cleaned up(no call to ClientClose or ServerClose\n");

  for (i=0;i<SSS_MAXPLEX;i++){
    if (sss->connections[i]){
      free(sss->connections[i]->handle);
      sss->connections[i]->close(sss->connections[i]->state);
      lt_dlclose(sss->connections[i]->libhandle);
      if (sss->connections[i]->host)
	free(sss->connections[i]->host);
      if (sss->connections[i]->protocol)
	free(sss->connections[i]->protocol);
      free(sss->connections[i]);
      sss->connections[i] = NULL;
    }
  }

  for (i=0;i<sss->numconfigs;i++){
    for (j=0;sss->configs[i].keywords[j];j++){
      free(sss->configs[i].keywords[j]);
      free(sss->configs[i].values[j]);
    }
    free(sss->configs[i].keywords);
    free(sss->configs[i].values);
  }

  free(sss->configs);

  if (sss->sd_host)
    free(sss->sd_host);
  if (sss->sd_schema_version)
    free(sss->sd_schema_version);
  if (sss->sd_tier)
    free(sss->sd_tier);

  if (sss->cluster)
    free(sss->cluster);

  if (sss->msglog)
    free(sss->msglog);

  for (i=0;sss->protoprefs[i];i++)
    free(sss->protoprefs[i]);

  free(sss->protoprefs);
  
  if (sss->servprotocol)
    free(sss->servprotocol);

  if (sss->sd_protocol)
    free(sss->sd_protocol);

  service_cache_free(sss->service_cache);

  free(sss);

  if (tempdebug)
    printf("ssslib.c: SSS_Close returning success\n");
  
  lt_dlexit();

  return 0;
}

int higher_priority_protocol(char **prefs,const char *first,const char *second){
  int i;
  
  if (!second)
    return 1;

  for (i=0;prefs[i];i++){
    if (!strcmp(prefs[i],first))
      return 1;
    if (!strcmp(prefs[i],second))
      return 0;
  }
  return 0;
}

void xml_starthandler(void *userData,const XML_Char *element_name,const XML_Char **attr){
  struct XML_Userdata *udata = userData;
  int i;

  struct SSS_Service service;

  switch (udata->messageType){
  case SSS_REG_RESP:
    if (!strcmp(element_name,"location"))
      udata->success = 1;
    break;
  case SSS_LOOKUP_RESP:
    if (!strcmp(element_name,"location")){
      udata->success = 1;
      for (i=0;attr[i];i += 2){
	if (!strcmp(attr[i],"host")){
	  service.host = (char *)attr[i+1];
	  udata->host = 1;
	}
	else if (!strcmp(attr[i],"port")){
	  service.extra = atoi((char *)attr[i+1]);
	  udata->extra = 1;
	}
	else if (!strcmp(attr[i],"protocol")){
	  service.protocol = (char *)attr[i+1];
	  udata->protocol = 1;
	}
	else if (!strcmp(attr[i],"component"))
	  service.service = (char *)attr[i+1];
	else if (!strcmp(attr[i],"schema_version"))
	  service.schema_version = (char *)attr[i+1];
	else if (!strcmp(attr[i],"tier"))
	  service.tier = (char *)attr[i+1];
      }
      if (udata->overwrite)
	service_cache_replace(udata->service_cache,&service);
      else
	service_cache_add(udata->service_cache,&service);
    }
    break;
  }
}

void xml_endhandler(void *userData,const XML_Char *element_name){

}

void SSS_strerror(int errnum,char *errbuf,int *type){
  switch (errnum){
  case SSS_XML_PARSE:
    *type = SSS_PARSE_ERROR;
    strncpy(errbuf,"Error parsing XML",MAXERRLENGTH);
    break;
  case SSS_NO_ROUTE:
    *type = SSS_CONNECT_ERROR;
    strncpy(errbuf,"No route to host",MAXERRLENGTH);
    break;
  case SSS_SD_FAIL:
    *type = SSS_INTERNAL_ERROR;
    strncpy(errbuf,"Error in communication with service directory",MAXERRLENGTH);
    break;
  case SSS_PROT_NAME_LONG:
    *type = SSS_PROTOCOL_ERROR;
    strncpy(errbuf,"Protocol name longer than 124 characters",MAXERRLENGTH);
    break;
  case SSS_NO_SUCH_CONN:
    *type = SSS_INTERNAL_ERROR;
    strncpy(errbuf,"A connection by that handle does not exist",MAXERRLENGTH);
    break;
  case SSS_PROTO_NO_SUPPORT:
    *type = SSS_SOCKET_ERROR;
    strncpy(errbuf,"Unsupported protocol",MAXERRLENGTH);
    break;
  case SSS_FAM_NO_SUPPORT:
    *type = SSS_SOCKET_ERROR;
    strncpy(errbuf,"Unsupported family",MAXERRLENGTH);
    break;
  case SSS_NOMEM:
    *type = SSS_INTERNAL_ERROR;
    strncpy(errbuf,"Out of memory",MAXERRLENGTH);
    break;
  case SSS_NOSOCKMEM:
    *type = SSS_SOCKET_ERROR;
    strncpy(errbuf,"No memory to make a new socket",MAXERRLENGTH);
    break;
  case SSS_INVALPROT:
    *type = SSS_SOCKET_ERROR;
    strncpy(errbuf,"Invalid protocol",MAXERRLENGTH);
    break;
  case SSS_HOST_NOT_FOUND:
    *type = SSS_CONNECT_ERROR;
    strncpy(errbuf,"Host not found",MAXERRLENGTH);
    break;
  case SSS_NO_ADDRESS:
    *type = SSS_CONNECT_ERROR;
    strncpy(errbuf,"No address given",MAXERRLENGTH);
    break;
  case SSS_UNRECOVERABLE_DNS_ERROR:
    *type = SSS_CONNECT_ERROR;
    strncpy(errbuf,"Unrecoverable DNS error",MAXERRLENGTH);
    break;
  case SSS_TRY_AGAIN:
    *type = SSS_CONNECT_ERROR;
    strncpy(errbuf,"Temporary DNS failure - try again",MAXERRLENGTH);
    break;
  case SSS_BAD_FD:
    *type = SSS_SOCKET_ERROR;
    strncpy(errbuf,"Bad file descriptor",MAXERRLENGTH);
    break;
  case SSS_FAULT:
    *type = SSS_INTERNAL_ERROR;
    strncpy(errbuf,"Memory fault",MAXERRLENGTH);
    break;
  case SSS_ALREADY_CONNECTED:
    *type = SSS_CONNECT_ERROR;
    strncpy(errbuf,"Socket is already connected",MAXERRLENGTH);
    break;
  case SSS_CONN_REFUSED:
    *type = SSS_CONNECT_ERROR;
    strncpy(errbuf,"Connection refused",MAXERRLENGTH);
    break;
  case SSS_TIMEOUT:
    *type = SSS_COMMUNICATION_ERROR;
    strncpy(errbuf,"Connection timed out",MAXERRLENGTH);
    break;
  case SSS_NET_UNREACH:
    *type = SSS_CONNECT_ERROR;
    strncpy(errbuf,"Network unreachable",MAXERRLENGTH);
    break;
  case SSS_ADDR_IN_USE:
    *type = SSS_SOCKET_ERROR;
    strncpy(errbuf,"Address already in use",MAXERRLENGTH);
    break;
  case SSS_NO_FREE_PORTS:
    *type = SSS_SOCKET_ERROR;
    strncpy(errbuf,"No free ports",MAXERRLENGTH);
    break;
  case SSS_NO_CONNECT_PERM:
    *type = SSS_CONNECT_ERROR;
    strncpy(errbuf,"No permissions to connect",MAXERRLENGTH);
    break;
  case SSS_NOSOCKPERM:
    *type = SSS_SOCKET_ERROR;
    strncpy(errbuf,"No permissions to make socket",MAXERRLENGTH);
    break;
  case SSS_NOT_A_SOCKET:
    *type = SSS_SOCKET_ERROR;
    strncpy(errbuf,"FD passed in is not a socket",MAXERRLENGTH);
    break;
  case SSS_SOCK_NOT_CONNECTED:
    *type = SSS_SOCKET_ERROR;
    strncpy(errbuf,"Socket is not connected",MAXERRLENGTH);
    break;
  case SSS_INVALID_SIZE:
    *type = SSS_INTERNAL_ERROR;
    strncpy(errbuf,"Invalid size",MAXERRLENGTH);
    break;
  case SSS_INTERRUPT:
    *type = SSS_INTERNAL_ERROR;
    strncpy(errbuf,"Interrupt",MAXERRLENGTH);
    break;
  case SSS_INVAL_ARG:
    *type = SSS_INTERNAL_ERROR;
    strncpy(errbuf,"Invalid argument",MAXERRLENGTH);
    break;
  case SSS_BROKEN_PIPE:
    *type = SSS_COMMUNICATION_ERROR;
    strncpy(errbuf,"Broken pipe",MAXERRLENGTH);
    break;
  case SSS_AUTH_FAILED:
    *type = SSS_PROTOCOL_ERROR;
    strncpy(errbuf,"Authentication error",MAXERRLENGTH);
    break;
  case SSS_LIBRARY_ERROR:
    *type = SSS_PROTOCOL_ERROR;
    snprintf(errbuf,MAXERRLENGTH,"Library module error");
    break;
  case SSS_EOF:
    *type = SSS_COMMUNICATION_ERROR;
    strncpy(errbuf,"EOF on read",MAXERRLENGTH);
    break;
  case SSS_PROT_ERROR:
    *type = SSS_PROTOCOL_ERROR;
    strncpy(errbuf,"Internal error in protocol",MAXERRLENGTH);
    break;
  case SSS_NEED_MORE_INFO:
    *type = SSS_PROTOCOL_ERROR;
    strncpy(errbuf,"Protocol needs more configuration information",MAXERRLENGTH);
    break;
  case SSS_RECV_TIMEOUT:
    *type = SSS_COMMUNICATION_ERROR;
    strncpy(errbuf,"Message receive timed out",MAXERRLENGTH);
    break;
  case SSS_CONN_RESET:
    *type = SSS_COMMUNICATION_ERROR;
    strncpy(errbuf,"Connection reset by peer",MAXERRLENGTH);
    break;
  case SSS_ILLEGAL_SEEK:
    *type = SSS_COMMUNICATION_ERROR;
    strncpy(errbuf,"Illegal seek",MAXERRLENGTH);
    break;
  case SSS_SD_LOOKUP_ERROR:
    *type = SSS_LOOKUP_ERROR;
    strncpy(errbuf,"No such component",MAXERRLENGTH);
    break;
  case SSS_PERMISSION_DENIED:
    *type = SSS_INTERNAL_ERROR;
    strncpy(errbuf,"Permission denied",MAXERRLENGTH);
    break;
  case SSS_NAMETOOLONG:
    *type = SSS_INTERNAL_ERROR;
    strncpy(errbuf,"File name too long",MAXERRLENGTH);
    break;
  case SSS_FILENOTFOUND:
    *type = SSS_INTERNAL_ERROR;
    strncpy(errbuf,"File not found",MAXERRLENGTH);
    break;
  default:
    *type = SSS_UNKNOWN_ERROR;
    strncpy(errbuf,"Unknown internal error",MAXERRLENGTH);
    break;
  }
}
  
void SSS_perror(SSS sss, char *prefix){
  char errbuf[MAXERRLENGTH+1];
  int type;
  
  SSS_strerror(sss->sss_errno,errbuf,&type);

  printf("%s: %s\n",prefix,errbuf);

}

char *SSS_GetFromConfigDict(SSS_ConfigDict dictionary, char *keyword){
  int i;
  for (i=0;dictionary->keywords[i];i++)
    if (!strcmp(dictionary->keywords[i],keyword))
      return dictionary->values[i];
  return NULL;
}

SSS_ConfigDict SSS_GetGlobalConfigDict(SSS sss){
  int port_len;
  char *port_buf;
  char snbuf[1];
  int dictlen;
  SSS_ConfigDict retptr = malloc(sizeof(struct SSS_ConfigDict_t));
  dictlen = sss->sd_host ? 7 : 2;
  retptr->protocol[0] = '\0';
  retptr->keywords = malloc(dictlen * sizeof(char *));
  retptr->values = malloc(dictlen * sizeof(char *));

  retptr->keywords[0] = my_strdup("debug");
  if (dictlen > 2){
    retptr->keywords[1] = my_strdup("sd_host");
    retptr->keywords[2] = my_strdup("sd_port");
    retptr->keywords[3] = my_strdup("sd_protocol");
    retptr->keywords[4] = my_strdup("sd_schema_version");
    retptr->keywords[5] = my_strdup("sd_tier");
    retptr->keywords[6] = NULL;
  }
  else
    retptr->keywords[1] = NULL;

  retptr->values[0] = my_strdup(sss->debug?"1":"0");
  if (dictlen > 2){
    retptr->values[1] = my_strdup(sss->sd_host);
    port_len = snprintf(snbuf,1,"%i",sss->sd_extra);
    port_buf = malloc(port_len+1);
    sprintf(port_buf,"%i",sss->sd_extra);
    retptr->values[2] = port_buf;
    retptr->values[3] = my_strdup(sss->sd_protocol);
    retptr->values[4] = my_strdup(sss->sd_schema_version);
    retptr->values[5] = my_strdup(sss->sd_tier);
    retptr->values[6] = NULL;
  }
  else
    retptr->values[1] = NULL;

  return retptr;
}

void SSS_FreeConfigDict(SSS_ConfigDict dict){
  int i;

  for(i=0;dict->keywords[i];i++){
    free(dict->keywords[i]);
    free(dict->values[i]);
  }
  free(dict->keywords);
  free(dict->values);
  free(dict);
}

